/*
 * JBoss, Home of Professional Open Source.
 * Copyright 2013, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags. See the copyright.txt file in the
 * distribution for a full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

package org.jboss.appendlog;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.ArrayDeque;

/**
 * // TODO: Document this
 *
 * @author Radim Vansa &lt;rvansa@redhat.com&gt;
 */
public abstract class SingleWriterAppendLog extends AbstractAppendLog {
   private ArrayDeque<ByteBuffer> queue = new ArrayDeque<ByteBuffer>(100);
   protected volatile long popIndex;
   private long pushIndex;
   private final Thread writer;
   private boolean run = true;
   private final int limit;

   public SingleWriterAppendLog(String filename, final int limit) throws FileNotFoundException {
      super(filename);
      this.limit = limit;
      writer = new Thread() {
         @Override
         public void run() {
            long myPopIndex = 0;
            int writtenNotFlushed = 0;
            while (true) {
               ByteBuffer buffer;
               synchronized (queue) {
                  buffer = queue.poll();
               }
               if (buffer == null) {
                  try {
                     flush();
                  } catch (IOException e) {
                     e.printStackTrace();
                     break;
                  }
                  writtenNotFlushed = 0;
                  setPopIndex(myPopIndex);
                  synchronized (queue) {
                     if (!run) return;
                     if (queue.isEmpty()) {
                        try {
                           queue.wait();
                        } catch (InterruptedException e) {
                        }
                     }
                  }
               } else {
                  try {
                     fileChannel.write(buffer);
                  } catch (IOException e) {
                     e.printStackTrace();
                     break;
                  }
                  myPopIndex++;
                  writtenNotFlushed++;
                  if (writtenNotFlushed >= limit) {
                     try {
                        flush();
                     } catch (IOException e) {
                        e.printStackTrace();
                        break;
                     }
                     setPopIndex(myPopIndex);
                     writtenNotFlushed = 0;
                  }
               }
            }
         }
      };
      writer.start();
   }

   @Override
   public void append(ByteBuffer bytes) throws IOException {
      long myIndex;
      synchronized (queue) {
         queue.push(bytes);
         queue.notify();
         pushIndex++;
         myIndex = pushIndex;
      }
      waitFor(myIndex);
   }

   protected abstract void waitFor(long myIndex);

   @Override
   public void close() throws IOException {
      synchronized (queue) {
         run = false;
         queue.notify();
      }
      try {
         writer.join();
      } catch (InterruptedException e) {
         e.printStackTrace();
      }
      fileChannel.close();
   }

   protected void setPopIndex(long popIndex) {
      this.popIndex = popIndex;
   }

   @Override
   public String toString() {
      return "SingleWriterAppendLog." + getClass().getSimpleName() + "{limit=" + limit + "}";
   }

   public static class Spinning extends SingleWriterAppendLog {
      public Spinning(String filename, int limit) throws FileNotFoundException {
         super(filename, limit);
      }

      @Override
      protected void waitFor(long myIndex) {
         while (myIndex > popIndex);
      }
   }

   public static class Yielding extends SingleWriterAppendLog {

      public Yielding(String filename, int limit) throws FileNotFoundException {
         super(filename, limit);
      }

      @Override
      protected void waitFor(long myIndex) {
         while (myIndex > popIndex) Thread.yield();
      }
   }

   public static class Waiting extends SingleWriterAppendLog {
      Object sync = new Object();

      public Waiting(String filename, int limit) throws FileNotFoundException {
         super(filename, limit);
      }

      @Override
      protected void waitFor(long myIndex) {
         synchronized (sync) {
            while (myIndex > popIndex) {
               try {
                  sync.wait();
               } catch (InterruptedException e) {
               }
            }
         }
      }

      @Override
      protected void setPopIndex(long popIndex) {
         this.popIndex = popIndex;
         synchronized (sync) {
            sync.notifyAll();
         }
      }
   }
}
