/*
 * JBoss, Home of Professional Open Source.
 * Copyright 2013, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags. See the copyright.txt file in the
 * distribution for a full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

package org.jboss.appendlog;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import junit.framework.TestCase;

/**
 * @author Radim Vansa &lt;rvansa@redhat.com&gt;
 */
public class AppendLogTest extends TestCase {

   private static final String testFile = System.getProperty("file", "/tmp/test");
   private static final long period = Long.getLong("period", 30000);
   private static final int size = Integer.getInteger("size", 1024);
   private static final int threadCount = Integer.getInteger("threads", -1);
   private static final int[] threadCounts = new int[] {4, 8, 12, 16, 20};

   private static volatile boolean start = false;
   private static volatile boolean stop = false;

   private static int[] getThreadCounts() {
      if (threadCount <= 0) return threadCounts;
      else return new int[] { threadCount };
   }

   public static void testSimple() throws FileNotFoundException {
      for (int threads : getThreadCounts()) {
         testWith(new SimpleAppendLog(testFile), threads);
      }
   }

   public static void testWaiting() throws FileNotFoundException {
      for (int threads : getThreadCounts()) {
         testWith(new WaitingAppendLog(testFile, 1, threads), threads);
         testWith(new WaitingAppendLog(testFile, 1, 2*threads), threads);
         testWith(new WaitingAppendLog(testFile, 1, threads/2), threads);
         testWith(new WaitingAppendLog(testFile, 2, threads), threads);
         testWith(new WaitingAppendLog(testFile, 2, 2*threads), threads);
         testWith(new WaitingAppendLog(testFile, 2, threads/2), threads);
         testWith(new WaitingAppendLog(testFile, 4, threads), threads);
         testWith(new WaitingAppendLog(testFile, 4, 2*threads), threads);
         testWith(new WaitingAppendLog(testFile, 4, threads/2), threads);
      }
   }

   public static void testSingleThread() throws FileNotFoundException {
      for (int threads : getThreadCounts()) {
         testWith(new SingleWriterAppendLog.Spinning(testFile, threads), threads);
         testWith(new SingleWriterAppendLog.Yielding(testFile, threads), threads);
         testWith(new SingleWriterAppendLog.Waiting(testFile, threads), threads);
      }
   }

   private static void testWith(AppendLog appendLog, int numThreads) {
      start = false;
      stop = false;
      List<WriterThread> threads = new ArrayList<WriterThread>(numThreads);
      for (int i = 0; i < numThreads; ++i) {
         WriterThread thread = new WriterThread("Writer-" + i, appendLog);
         threads.add(thread);
         thread.start();
      }
      long startTime = System.nanoTime();
      start = true;
      try {
         Thread.sleep(period);
      } catch (InterruptedException e) {
         e.printStackTrace();
      }
      stop = true;
      long sum = 0;
      for (WriterThread thread : threads) {
         try {
            thread.join();
         } catch (InterruptedException e) {
            e.printStackTrace();
         }
         sum += thread.counter;
      }
      long endTime = System.nanoTime();
      try {
         appendLog.close();
      } catch (IOException e) {
         e.printStackTrace();
      }
      System.out.printf("%2d threads, %9d appends (%9d flushes) in %5.2f ms to %s\n",
            numThreads, sum, appendLog.getFlushCount(), (endTime - startTime) / 1000000d, appendLog);
   }

   private static class WriterThread extends Thread {
      private final ByteBuffer buffer;
      private final AppendLog appendLog;
      private long counter = 0;

      private WriterThread(String name, AppendLog appendLog) {
         super(name);
         byte[] bytes = new byte[size];
         new Random().nextBytes(bytes);
         buffer = ByteBuffer.wrap(bytes);
         this.appendLog = appendLog;
      }

      @Override
      public void run() {
         while (!start) Thread.yield();
         while (!stop) {
            try {
               appendLog.append(buffer);
               buffer.rewind();
               counter++;
            } catch (IOException e) {
               e.printStackTrace();
               break;
            }
         }
      }
   }
}
